//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis                         
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-                                
//*  http://apps.jcns.fz-juelich.de/frida                                   
//**************************************************************************************************

//! \file  expr.cpp
//! \brief Parse expressions using CRef, CContext, CObj, virtual CNode

#include <string>
#include <vector>
#include <algorithm>

#include "../trivia/vector_ops.hpp"

#include "defs.hpp"
#include "olf.hpp"
#include "mem.hpp"
#include "slice.hpp"
#include "ptr.hpp"
#include "obj.hpp"
#include "expr.hpp"
#include "xax_lex.hpp"


//**************************************************************************************************
//*  CContext                                                               
//**************************************************************************************************


//! Constructor. Sets k,j,i to arguments. Sets other members to defaults.

CContext::CContext( int _k, int _j, int _i )
    : k(_k), j(_j), i(_i),                     // set from arguments
      vt(0), dim(_1), want_error(false), nv(0) // can be changed later
{
    if ( k==(int)-1 )
        return; // not sure whether k=-1 makes sense
    POlo f = NOlm::mem_get(k);
    if ( j==(int)-1 ) {
        if ( f->nJ()==1 )
            j = 0;
    } else if ( j>=f->nJ() )
        throw "invalid evaluation context, j="+S(j)+">=nJ="+S(f->nJ());

    POld fd = P2D(f);
    if ( !fd ) {
        if ( i!=(int)-1 )
            throw S("invalid evaluation context, i set for curve file");
        return;
    }

    if ( j==(int)-1 )
        return;
    int ni = fd->nPts(j);
    if ( i==(int)-1 ) {
        if ( ni==1 )
            i = 0;
    } else if ( i>=ni )
        throw "invalid evaluation context for j="+S(j)+": i="+
            S(i)+">=ni="+S(ni);        
}


//! Sets request of a vector of doubles, corresponding to input points 't'.

void CContext::request_VT( const vector<double> *_vt )
{
    dim = _VT;
    nv = _vt->size();
    vt = _vt;
    if( !triv::is_equidist( &vt_step, *vt ) || vt_step<0 )
        vt_step = 0;
}

//! Sets request of a vector of _nv indices.

void CContext::request_VI( int _nv )
{
    dim = _VI;
    nv = _nv;
}

//! Returns textual representation of the class instance.

string CContext::context_info() const
{
    string out = "";
    if      ( dim==_1 )
        out += "dim 1, ";
    else if ( dim==_VI )
        out += "dimVI, ";
    else if ( dim==_VT )
        out += "dimVT, ";
    out += "k = " + S(k) + ", ";
    out += "j = " + S(j) + ", ";
    out += "i = " + S(i) + ", ";
    out += "nv = " + S(nv) + ", ";
    if( vt )
        out += "size(vt) = " + S(vt->size()) + ", ";
    else
        out += "no vt, ";
    out += "want_error = " + S(want_error);
    return out;
}


//**************************************************************************************************
//*  CRef (references like y[k,j,0] or p3[2] or ...)                        
//**************************************************************************************************


//! Returns file index k in given context.

int CRef::get_k( const CContext& ctx ) const
{
    int k;
    if ( tk ) {
        RObj tmp = tk->tree_val( ctx );
        RObjInt pi = PCAST<const CObjInt>(tmp);
        if ( !pi )
            throw "index K=" + tmp->to_s() + " is not integer";
        k = pi->val;
    } else {
        k = ctx.k;
    }
    if ( k==(int)-1 )
        throw S("missing or invalid k");
    NOlm::mem_get( k ); // to assert existence of file
    return k;
}


//! Returns spectrum index j in given context.

int CRef::get_j( const CContext& ctx, int nj ) const
{
    int j;
    if ( tj ) {
        RObj tmp = tj->tree_val( ctx );
        RObjInt pi = PCAST<const CObjInt>(tmp);
        if ( !pi )
            throw "index J=" + tmp->to_s() + " is not integer";
        j = pi->val;
    } else {
        j = ctx.j;
    }
    if ( j==-1 )
        throw S("index J=-1 (not defined?)");
    else if ( j<0 || j>=nj )
        throw "index J=" + S(j) + " is out of bounds";
    return j;
}

//! Returns string representation of the class instance.

string CRef::ref_info() const
{
    string out = "ref: ";
    if ( tk )
        out += "tk[" + tk->tree_info() + "], ";
    else
        out += "tk=0, ";
    if ( tj )
        out += "tj[" + tj->tree_info() + "], ";
    else
        out += "tj=0, ";
    if ( ti )
        out += "ti[" + ti->tree_info() + "]";
    else
        out += "ti=0";
    return out;
}

//**************************************************************************************************
//*  CNode: status requests                                                 
//**************************************************************************************************


//! Returns number of curve parameters used in node and below.

int CNode::npar() const
{
    int np=0;
    npar_exec( &np ); // implemented with pointer argument to allow recursion
    return np;
}


//! Does the node depend on k?

bool CNode::k_dependent() const
{
    bool kd=false;
    kdep_exec( &kd );
    return kd;
}


//**************************************************************************************************
//*  CNode: low-level tree evaluation                                       
//**************************************************************************************************


//! Evaluates the node, and returns a non-negative integer.

int CNode::tree_val_idx( const CContext& ctx, const string& name ) const
{
    RObj val = tree_val( ctx );
    RObjInt pi = PCAST<const CObjInt>(val);
    if ( !pi )
        throw "Index "+name+"="+val->to_s()+" is not integer";
    int ret = pi->val;
    if( ret<0 )
        throw "Index "+name+"="+val->to_s()+" is negativ";
    return ret;
}


//! Evaluates the node, and returns a list of non-negative integers.

RObjVecInt CNode::to_index_list( int siz ) const
{
    CContext ctx;
    int nret = siz;
    ctx.request_VI( nret );
    ctx.want_error = false;
    RObj val = tree_val( ctx );
    if ( RObjInt ri = PCAST<const CObjInt>( val ) ) {
        PObjVecInt ret( new CObjVecInt );
        int ival = ri->to_i();
        if( ival<0 || ival>=siz )
            return NULL;
        ret->v.push_back( ival );
        return ret;
    } else if ( RObjVecInt riv = PCAST<const CObjVecInt>( val ) ) {
        for( int ival: riv->v )
            if( ival<0 || ival>=siz )
                return NULL;
        return riv;
    } else {
        return NULL;
    }
}

//**************************************************************************************************
//*  CNode: public interface to tree evaluation                             
//**************************************************************************************************

//! Evaluates the node in a scalar context.

RObj CNode::tree_point_val( int k, int j, int i ) const
{
    CContext ctx( k, j, i );
    ctx.want_error = true;
    return tree_val( ctx );
}


//! Evaluates the node, and returns a double. OBSOLET!?

double CNode::tree_point_dbl( int k, int j, int i ) const
{
    CContext ctx( k, j, i );
    ctx.want_error = false;
    RObj val = tree_val( ctx );
    RObjNum pr = PCAST<const CObjNum>(val);
    if ( !pr )
        throw S("Tree point value has unexpected type");
    return pr->to_r();
}


//! Evaluates the node, and sets vectors *ret and *dret according to the result.

void CNode::tree_vec_val( vector<double> *ret, vector<double> *dret, int k, int j ) const
{
    CContext ctx( k, j );
    int nret = ret->size();
    ctx.request_VI( nret );
    ctx.want_error = (bool)dret;
    RObj val = tree_val( ctx );
    if        ( RObjNum pr = PCAST<const CObjNum>(val) ) {
        ret->assign( nret, pr->to_r() );
        if( dret && pr->has_err() )
            dret->assign( nret, pr->to_dr() );

    } else if ( RObjVecInt pv = PCAST<const CObjVecInt>(val) ) {
        for( int i=0; i<nret; ++i )
            (*ret)[i] = pv->to_r(i);
    } else if ( RObjVecDbl pv = PCAST<const CObjVecDbl>(val) ) {
        *ret = pv->v;
        if( dret ) {
            if ( RObjVecEnu pdv = PCAST<const CObjVecEnu>(val) )
                *dret = pdv->dv;
            else
                dret->clear();
        }
    } else
        throw S("BUG in tree_vec_val: unexpected return type");
}

//! Returns a node obtained by parsing the given expression.

RObj CNode::eval( string expr )
{
    RNode T = user_xaxparse( expr.c_str() );
    if( T->k_dependent() )
        throw "Cannot evaluate '" + expr + "': missing context K";
    return T->tree_point_val();
}
