//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  file_in.cpp
//! \brief NFileIn: load files.

#include <cstdlib>
#include <cstring>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <map>
#include <yaml-cpp/yaml.h>

#include "../trivia/string_ops.hpp"
#include "../trivia/vector_ops.hpp"
#include "../trivia/file_ops.hpp"
#include "../readplus/ask.hpp"

#include "defs.hpp"
#include "olf.hpp"
#include "mem.hpp"
#include "var.hpp"
#include "slice.hpp"
#include "obj.hpp"
#include "expr.hpp"
#include "file_in.hpp"

using namespace std;

namespace NFileIn {
    void Load_08( ifstream& F_in, string flong );
}


//! Load file in current or legacy format.

void NFileIn::load(void)
{
    string fnames = sask( "Load file(s)" );
    vector<string> vflong; // unix names
    triv::glob_file_list( fnames, "y08", &vflong );
    for( int i=0; i<vflong.size(); ++i ) {
        FILE *F_in = fopen(vflong[i].c_str(), "r");
        if( !F_in )
            throw "cannot open file " + vflong[i];
        try{
            cout << ".. loading file " << vflong[i] << "\n";
            string fdir, fshort, fext;
            triv::fname_divide( vflong[i], &fdir, &fshort, &fext);

            if        ( fext=="y08" ) {
                ifstream FS (vflong[i].c_str(), ios_base::in);
                if( FS.fail() )
                    throw S("cannot open file");
                Load_08( FS, vflong[i] );
                FS.close();
            } else
                throw "unknown file extension [" + fext + "]";
            fclose( F_in );
        }
        catch(string& ex) {
            throw "Cannot load " + vflong[i] + ": " + ex;
        }
        catch(std::exception& ex) {
            throw "Cannot load " + vflong[i] + ": " + ex.what();
        }
    }
}

//! Load a YAML file in y08 format.

void NFileIn::Load_08(ifstream& FS,string flong)
{
    string fdir, fshort, fext;
    triv::fname_divide( flong, &fdir, &fshort, &fext);

    const YAML::Node doc = YAML::Load(FS);
    if( doc.Type() != YAML::NodeType::Map )
        throw S("document root is not of MAP type");

    //read meta
    if(!doc["Meta"])
        throw S("no Meta");
    if ( doc["Meta"].Type() != YAML::NodeType::Map && doc["Meta"].size() > 0 )
        throw S("Meta is not a MAP type");

    if(!doc["Meta"]["format"])
        throw S("no format in Meta");
    string s = doc["Meta"]["format"].as<string>();
    if ( s != "frida/y08 for yaml1" )
        throw S("format is not frida/y08 for yaml1");

    if(!doc["Meta"]["type"])
        throw S("no type in Meta");
    string type = doc["Meta"]["type"].as<string>();

    // create output file
    POlo fout;
    POld fd;
    POlc fc;
    bool isdata;
    if ( type == "generic tabular data" ){
        isdata = true;
        fd = POld( new COld );
        fout = fd;
    } else if ( type == "frida2 curve" ){
        isdata = false;
        fc = POlc( new COlc );
        fout = fc;
    } else
        throw "File "+flong+" has invalid type "+type;

    // read history
    if(!doc["History"])
        throw S("no History");
    if ( doc["History"].Type() != YAML::NodeType::Sequence  && doc["History"].size() > 0 )
        throw S("History is not a SEQUENCE type");
    for (int i = 0; i < doc["History"].size(); i++) {
        string s = doc["History"][i].as<string>();
        fout->lDoc.push_back(s);
    }

    // read coord
    if(!doc["Coord"])
        throw S("no Coord");
    if ( doc["Coord"].Type() != YAML::NodeType::Map && doc["Coord"].size() > 0 )
        throw S("Coord is not a MAP type");
    if( const YAML::Node& pName = doc["Coord"]["x"] ){
        fout->xco.name = pName["name"].as<string>();
        fout->xco.unit =pName["unit"].as<string>();
    } else
        throw S("no x coord");
    if( const YAML::Node& pName = doc["Coord"]["y"] ){
        fout->yco.name = pName["name"].as<string>();
        fout->yco.unit = pName["unit"].as<string>();
    } else
        throw S("no y coord");

    int iz = 0;
    while(const YAML::Node& pName = doc["Coord"]["z"+S(iz)]){
        CCoord co(pName["name"].as<string>(), pName["unit"].as<string>());
        fout->ZCo.push_back( co );
        ++iz;
    }

    // read Param
    if(!doc["Param"])
        throw S("no Param");
    if ( doc["Param"].Type() != YAML::NodeType::Sequence  && doc["Param"].size() > 0)
        throw S("Param is not a SEQUENCE type");
    for( int iParam = 0; iParam < (doc["Param"].size()); iParam++ ){
        if ( doc["Param"][iParam].Type() != YAML::NodeType::Map )
            throw "Param " + S(iParam) + " is not a MAP type";
        CCoord co(doc["Param"][iParam]["name"].as<string>(),
                  doc["Param"][iParam]["unit"].as<string>());
        string s = doc["Param"][iParam]["value"].as<string>();
        double num;
        if( !triv::any2dbl( s, &num ) )
            throw "param(" + co.name + "): invalid value " + s;
        fout->RPar.push_back( CParam( co, num ) );
    }

    // for curve, read formula
    if( !isdata ) {
        const YAML::Node& pName = doc["Formula"];
        if( !pName )
            throw S("DEFICIENT FILE: no formula");
        string s = pName.as<string>();
        fc->parse_function( s );
        fc->curve_set_defaults();
    }

    //start to read Tables
    if(!doc["Tables"])
        throw S("no Tables");
    if ( doc["Tables"].Type() != YAML::NodeType::Sequence && doc["Tables"].size() > 0 )
        throw S("Tables is not a SEQUENCE type");
    for (int iTable = 0; iTable < (doc["Tables"].size()); iTable++) {
        if ( doc["Tables"][iTable].Type() != YAML::NodeType::Map )
            throw "Tables " + S(iTable) + " is not a MAP type";
        vector<RObj> z;
        for(int iz=0; iz<fout->ZCo.size(); ++iz ){
            string s = doc["Tables"][iTable]["z" + S(iz)].as<string>() ;
            double num;
            if( !triv::any2dbl( s, &num ) )
                throw "z" + S(iz) + ": invalid value " + s;
            z.push_back( RObjDbl( new CObjDbl( num ) ) );
        }

        if( isdata ) {
            PSpec sout( new CSpec );
            sout->z = z;
            string xytag;
            if ( doc["Tables"][iTable]["xy"] ) {
                string res, lin;
                string val = doc["Tables"][iTable]["xy"].as<string>();
                while( val != "" ) {
                    triv::string_extract_line( val, &lin, &res );
                    double vx, vy;
                    if( sscanf( lin.c_str(), "%lg %lg\n", &vx, &vy )!=2 )
                        throw "xy: bad data line: " + lin;
                    sout->push_xy( vx, vy );
                    val = res;
                };
            } else if ( doc["Tables"][iTable]["xyd"] ) {
                //itxy.second() >> val;
                string res, lin;
                string val = doc["Tables"][iTable]["xyd"].as<string>();
                while( val != "" ) {
                    triv::string_extract_line( val, &lin, &res );
                    double vx, vy, dy;
                    if( sscanf( lin.c_str(), "%lg %lg %lg\n",&vx, &vy, &dy )!=3 )
                        throw "xyd: bad data line: " + lin;
                    sout->push_xyd( vx, vy, dy );
                    val = res;
                };
            } else
                throw "invalid xytag " + xytag;
            fout->V.push_back( sout );
        } else {
            PCurve cout( new CCurve );
            cout->z = z;
            for( int ip=0; ip<fc->nP; ++ip ){
                string res;
                string val = doc["Tables"][iTable]["p" + S(ip)].as<string>();
                double num;
                if( !triv::any2dbl( val, &num ) )
                    throw "p" + S(ip) + ": invalid value " + val;
                cout->P.push_back( num );
                cout->ParAttr.push_back( 'u' );
            }
            fout->V.push_back( cout );
        }
    }

    fout->name = fshort;
    fout->as_on_disk = true;
    NOlm::mem_store( fout );
}
