//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-, this file 2009-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  file_out.cpp
//! \brief NFileOut: write files.

#include <cmath>
#include <string>
#include <vector>

#include "../trivia/yaml_out.hpp"
#include "../readplus/ask.hpp"

#include "defs.hpp"
#include "olf.hpp"
#include "mem.hpp"
#include "obj.hpp"
#include "slice.hpp"
#include "file_out.hpp"

namespace NFileOut {
    void save_y08( FILE *file, POlo f );
    void save_csv( FILE *file, POlo f );
    void save_tab( FILE *file, POlo f );
}


//! Queries file names, and writes files.

void NFileOut::save( string fmt, bool allow_overwrite )
{
    string outfnam;
    FILE *file;
    NOlm::IterateO iter;
    POlo f;
    while( (f=iter()) ) {
        // query output file name
        outfnam = wask( "Save as (."+fmt+")", f->name);
        if ( outfnam=="" )
            return;
        f->name = outfnam;
        outfnam += "."+fmt;
        // document file save
        f->lDoc.push_back( "fs " + outfnam + " # " + triv::time2strg( time(0) ) );
        // prevent unintended overwriting
        if( !allow_overwrite ){
            if ( (file = fopen(outfnam.c_str(), "r")) ) {
                fclose(file);
                throw "file "+outfnam+" exists, use fso to overwrite";
            }
        }
        // save file
        if ( !(file = fopen(outfnam.c_str(), "w")) )
            throw "cannot write to file " + outfnam;
        if( fmt=="y08"){
            save_y08( file, f );
        } else if( fmt=="csv" ) {
            save_csv( file, f );
        } else if( fmt=="tab" ) {
            save_tab( file, f );
        } else
            throw "Unknown format "+fmt;
        fclose(file);
        f->as_on_disk = true;
    }
}


//! Writes a file in format y08.

void NFileOut::save_y08( FILE *file, POlo f )
{
    POld fd = P2D( f );
    POlc fc = P2C( f );

    fprintf( file, "Meta:\n"
             "  format:  frida/y08 for yaml1\n"
             "  type:    %s\n",
             fd ? "generic tabular data" : "frida2 curve" );

    fprintf( file, "History:\n" );
    for ( string lin: f->lDoc )
        fprintf( file, "- %s\n", triv::yaml(lin).c_str() );

    fprintf( file, "Coord:\n" );
    fprintf( file, "  x:\n    name: %s\n    unit: %s\n",
             triv::yaml(f->xco.name).c_str(), triv::yaml(f->xco.unit).c_str() );
    fprintf( file, "  y:\n    name: %s\n    unit: %s\n",
             triv::yaml(f->yco.name).c_str(), triv::yaml(f->yco.unit).c_str() );
    for ( int i=0; i<f->ZCo.size(); i++ )
        fprintf( file, "  z%i:\n    name: %s\n    unit: %s\n",
                 i, triv::yaml(f->ZCo[i].name).c_str(),
                 triv::yaml(f->ZCo[i].unit).c_str() );

    fprintf( file, "Param:\n" );
    for ( int i=0; i<f->RPar.size(); i++ )
        fprintf( file,
                 "  -\n    name:  %s\n    unit:  %s\n    value: %20.15g\n",
                 triv::yaml(f->RPar[i].Co.name).c_str(),
                 triv::yaml(f->RPar[i].Co.unit).c_str(),
                 f->RPar[i].val );

    if( fc ) {
        fprintf( file, "Formula: %s\n", fc->expr.c_str() );
    }

    fprintf( file, "Tables: # %i tables follow\n", f->nJ() );

    for( int j=0; j<f->nJ(); j++ ){
        fprintf( file, "  - # table %i\n", j );
        PSlice s = f->V[j];
        for ( int i=0; i<f->nZ(); i++ ) {
            if( RObjNum pz = PCAST<const CObjNum>(s->z[i]) ) {
                fprintf( file, "    z%i: %18.10g\n", i, pz->to_r() );
                if( RObjNum pdz = PCAST<const CObjEnu>(s->z[i]) )
                    fprintf( file, "    dz%i: %18.10g\n", i, pdz->to_dr() );
            } else
                fprintf( file, "    z%i: %s\n", i, s->z[i]->to_s().c_str() );
        }
        if   ( fd ) {
            PSpec s = fd->VS(j);
            if( s->dy.size() ) {
                fprintf( file, "    xyd: |2 # %i entries\n", (int)s->size() );
                for( int i=0; i<s->size(); i++ )
                    fprintf( file, "      %20.15g %20.15g %20.15g\n",
                             s->x[i], s->y[i], s->dy[i] );
            } else {
                fprintf( file, "    xy: |2 # %i entries\n", (int)s->size() );
                for( int i=0; i<s->size(); i++ )
                    fprintf( file, "      %20.15g %20.15g\n",
                             s->x[i], s->y[i] );
            }
        } else {
            PCurve s = fc->VC(j);
            for ( int i=0; i<fc->nP; i++ )
                fprintf( file, "    p%i: %18.10g\n", i, s->P[i] );
        }
    }
}


//! Writes y as tab-separated table.

void NFileOut::save_csv( FILE *file, POlo f )
{
    POld fd = P2D( f );
    POlc fc = P2C( f );
    if( !fd )
        throw S("csv save not implemented for non-data files");
    for( int j=0; j<f->nJ(); j++ ){
        for( int i=0; i<fd->VS(j)->size(); ++i ){
            fprintf( file, "%g\t", fd->VS(j)->y[i] );
        }
        fprintf( file, "\n" );
    }
}


//! Writes data as x-y columns with z header lines.

void NFileOut::save_tab( FILE *file, POlo f )
{
    POld fd = P2D( f );
    POlc fc = P2C( f );
    if( !fd )
        throw S("tab save not implemented for non-data files");
    for( int j=0; j<f->nJ(); j++ ){
        for( int iz=0; iz<f->ZCo.size(); ++iz ){
            if( iz>0 )  fprintf( file, "\t" );
            fprintf( file, "%g", PCAST<const CObjNum>(fd->VS(j)->z[iz])->to_r() );
        }
        fprintf( file, "\n" );
        for( int i=0; i<fd->VS(j)->size(); ++i ){
            fprintf( file, "%g\t%g\n", fd->VS(j)->x[i], fd->VS(j)->y[i] );
        }
        fprintf( file, "\n" );
    }
}
