//**************************************************************************************************
//* FRIDA: fast reliable interactive data analysis
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-
//* http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  node.cpp
//! \brief Nodes within expression trees, inheriting from virtual base CNode.


#include <map>
#include <algorithm>
#include <iostream>

#include "defs.hpp"
#include "olf.hpp"
#include "mem.hpp"
#include "slice.hpp"
#include "reg.hpp"
#include "func.hpp"
#include "fregistry.hpp"
#include "geni.hpp"
#include "integrate.hpp"
#include "var.hpp"
#include "obj.hpp"
#include "expr.hpp"
#include "node.hpp"
#include "curve.hpp"

extern bool allow_slow_conv; // set in curve.cpp

#define DEBUG 0


//**************************************************************************************************
//* CNodeWithArgs
//**************************************************************************************************

bool CNodeWithArgs::has_dummy() const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        if( arg[iarg] && arg[iarg]->has_dummy() )
            return true;
    return false;
}

bool CNodeWithArgs::has_conv() const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        if( arg[iarg] && arg[iarg]->has_conv() )
            return true;
    return false;
}

void CNodeWithArgs::kdep_exec( bool *kd ) const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        if( arg[iarg] )
            arg[iarg]->kdep_exec( kd );
}

void CNodeWithArgs::npar_exec( int *np ) const
{
    for ( int iarg=0; iarg<narg; ++iarg )
        if( arg[iarg] )
            arg[iarg]->npar_exec( np );
}


//**************************************************************************************************
//* CNodeFun: function/operator node
//**************************************************************************************************

CNodeFun1::CNodeFun1( const CFunc* _fu, RNode _a0 )
    : CNodeFun(1, _fu)
{
    arg[0] = _a0;
}

CNodeFun2::CNodeFun2( const CFunc* _fu, RNode _a0, RNode _a1 )
    : CNodeFun(2, _fu)
{
    arg[0] = _a0;
    arg[1] = _a1;
}

CNodeFun3::CNodeFun3( const CFunc* _fu, RNode _a0, RNode _a1, RNode _a2 )
    : CNodeFun(3, _fu)
{
    arg[0] = _a0;
    arg[1] = _a1;
    arg[2] = _a2;
}

//! Returns the result of function evaluation for given arguments.

const RObj CNodeFun::tree_val( const CContext& ctx ) const
{
    string basetypes;
    const CTypedFunc* tf = NULL;
    try {
         // evaluate arguments, determine if all results are scalar.
        vector<RObj> pa(narg);
        basetypes = "";
        bool is_scalar = true;
        for ( int iarg=0; iarg<narg; ++iarg ) {
            pa[iarg] = arg[iarg]->tree_val( ctx );
            basetypes += pa[iarg]->base_type();
            if ( pa[iarg]->is_vec() )
                is_scalar = false;
        }

        tf = fu->find_tyfu( basetypes, ctx.want_error );
        if( !tf )
            throw "function " + fu->tag + " not implemented for argument types " + basetypes;
        funcPtr f = tf->f;

        if( is_scalar ) {
            if        ( tf->outtype=="0" ) {
                if       ( tf->intypes=="i" )
                    (*(func_0_i)(f))( pa[0]->to_i() );
                else if  ( tf->intypes=="d" )
                    (*(func_0_d)(f))( pa[0]->to_r() );
                else if  ( tf->intypes=="s" )
                    (*(func_0_s)(f))( pa[0]->to_s() );
                else
                    throw S("BUG: unexpected intypes");
                return RObjStr( new CObjStr( "__NULL__" ) );
            } else if ( tf->outtype=="i" ) {
                int val;
                if       ( tf->intypes=="i" )
                    val = (*(func_i_i)(f))( pa[0]->to_i() );
                else if  ( tf->intypes=="d" )
                    val = (*(func_i_d)(f))( pa[0]->to_r() );
                else if  ( tf->intypes=="s" )
                    val = (*(func_i_s)(f))( pa[0]->to_s() );
                else if  ( tf->intypes=="ii" )
                    val = (*(func_i_ii)(f))( pa[0]->to_i(), pa[1]->to_i() );
                else if  ( tf->intypes=="is" )
                    val = (*(func_i_is)(f))( pa[0]->to_i(), pa[1]->to_s() );
                else if  ( tf->intypes=="dd" )
                    val = (*(func_i_dd)(f))( pa[0]->to_r(), pa[1]->to_r() );
                else if  ( tf->intypes=="si" )
                    val = (*(func_i_si)(f))( pa[0]->to_s(), pa[1]->to_i() );
                else if  ( tf->intypes=="iii" )
                    val = (*(func_i_iii)(f))( pa[0]->to_i(), pa[1]->to_i(), pa[2]->to_i() );
                else
                    throw S("BUG: unexpected intypes");
                return RObjInt( new CObjInt( val ) );

            } else if ( tf->outtype=="d" ) {
                double val;
                if       ( tf->intypes=="i" )
                    val = (*(func_d_i)(f))( pa[0]->to_i() );
                else if  ( tf->intypes=="d" )
                    val = (*(func_d_d)(f))( pa[0]->to_r() );
                else if  ( tf->intypes=="s" )
                    val = (*(func_d_s)(f))( pa[0]->to_s() );
                else if  ( tf->intypes=="ii" )
                    val = (*(func_d_ii)(f))( pa[0]->to_i(), pa[1]->to_i() );
                else if  ( tf->intypes=="dd" )
                    val = (*(func_d_dd)(f))( pa[0]->to_r(), pa[1]->to_r() );
                else if  ( tf->intypes=="sd" )
                    val = (*(func_d_sd)(f))( pa[0]->to_s(), pa[1]->to_r() );
                else if  ( tf->intypes=="ddd" )
                    val = (*(func_d_ddd)(f))( pa[0]->to_r(), pa[1]->to_r(), pa[2]->to_r() );
                else
                    throw S("BUG: unexpected intypes");
                return RObjDbl( new CObjDbl( val ) );
            } else if ( tf->outtype=="e" ) {
                double val, err;
                if       ( tf->intypes=="e" )
                    (*(func_e_e)(f))( val, err,
                             pa[0]->to_r(), pa[0]->to_dr() );
                else if  ( tf->intypes=="dd" )
                    (*(func_e_dd)(f))( val, err, pa[0]->to_r(), pa[1]->to_r() );
                else if  ( tf->intypes=="ee" )
                    (*(func_e_ee)(f))( val, err,
                             pa[0]->to_r(), pa[0]->to_dr(),
                             pa[1]->to_r(), pa[1]->to_dr() );
                else if  ( tf->intypes=="se" )
                    (*(func_e_se)(f))( val, err,
                             pa[0]->to_s(),
                             pa[1]->to_r(), pa[1]->to_dr() );
                else if  ( tf->intypes=="eee" )
                    (*(func_e_eee)(f))( val, err,
                             pa[0]->to_r(), pa[0]->to_dr(),
                             pa[1]->to_r(), pa[1]->to_dr(),
                             pa[2]->to_r(), pa[2]->to_dr() );
                else
                    throw S("BUG: unexpected intypes");
                return RObjEnu( new CObjEnu( val, err ) );
            } else if ( tf->outtype=="s" ) {
                string val;
                if       ( tf->intypes=="s" )
                    val = (*(func_s_s)(f))( pa[0]->to_s() );
                else if  ( tf->intypes=="si" )
                    val = (*(func_s_si)(f))( pa[0]->to_s(), pa[1]->to_i() );
                else if  ( tf->intypes=="ss" )
                    val = (*(func_s_ss)(f))( pa[0]->to_s(), pa[1]->to_s() );
                else
                    throw S("BUG: unexpected intypes");
                return PObjStr( new CObjStr( val ) );
            } else
                throw S("BUG: unexpected outtype");
        } else { //is_vec()
            // determine length of input vectors
            int n=1;
            for ( int iarg=0; iarg<narg; ++iarg ) {
                int na = pa[iarg]->size();
                if      ( n==1 )
                    n = na;
                else if ( na != 1 && na != n )
                    throw S("vector arguments have different sizes");
            }
            if        ( tf->outtype=="i" ) {
                PObjVecInt ret( new CObjVecInt( n ) );
                if       ( tf->intypes=="i" )
                    for ( int i=0; i<n; ++i )
                        ret->v[i] = (*(func_i_i)(f))( pa[0]->to_i(i) );
                else if  ( tf->intypes=="d" )
                    for ( int i=0; i<n; ++i )
                        ret->v[i] = (*(func_i_d)(f))( pa[0]->to_r(i) );
                else if  ( tf->intypes=="ii" )
                    for ( int i=0; i<n; ++i )
                        ret->v[i] = (*(func_i_ii)(f))( pa[0]->to_i(i), pa[1]->to_i(i) );
                else if  ( tf->intypes=="dd" )
                    for ( int i=0; i<n; ++i )
                        ret->v[i] = (*(func_i_dd)(f))( pa[0]->to_r(i), pa[1]->to_r(i) );
                else if  ( tf->intypes=="iii" )
                    for ( int i=0; i<n; ++i )
                        ret->v[i] = (*(func_i_iii)(f))( pa[0]->to_i(i), pa[1]->to_i(i), pa[2]->to_i(i) );
                else
                    throw S("BUG: unexpected intypes");
                return ret;
            } else if ( tf->outtype=="d" ) {
                PObjVecDbl ret( new CObjVecDbl( n ) );
                if       ( tf->intypes=="i" )
                    for ( int i=0; i<n; ++i )
                        ret->v[i] = (*(func_d_i)(f))( pa[0]->to_i(i) );
                else if  ( tf->intypes=="d" )
                    for ( int i=0; i<n; ++i )
                        ret->v[i] = (*(func_d_d)(f))( pa[0]->to_r(i) );
                else if  ( tf->intypes=="ii" )
                    for ( int i=0; i<n; ++i )
                        ret->v[i] = (*(func_d_ii)(f))( pa[0]->to_i(i), pa[1]->to_i(i) );
                else if  ( tf->intypes=="dd" )
                    for ( int i=0; i<n; ++i )
                        ret->v[i] = (*(func_d_dd)(f))( pa[0]->to_r(i), pa[1]->to_r(i) );
                else if  ( tf->intypes=="ddd" )
                    for ( int i=0; i<n; ++i )
                        ret->v[i] = (*(func_d_ddd)(f))( pa[0]->to_r(i), pa[1]->to_r(i), pa[2]->to_r(i) );
                else
                    throw S("BUG: unexpected intypes");
                return ret;
            } else if ( tf->outtype=="e" ) {
                PObjVecEnu ret( new CObjVecEnu( n ) );
                if       ( tf->intypes=="e" )
                    for ( int i=0; i<n; ++i )
                        (*(func_e_e)(f))( ret->v[i], ret->dv[i], pa[0]->to_r(i), pa[0]->to_dr(i) );
                else if  ( tf->intypes=="dd" )
                    for ( int i=0; i<n; ++i )
                        (*(func_e_dd)(f))( ret->v[i], ret->dv[i], pa[0]->to_r(i), pa[1]->to_r(i) );
                else if  ( tf->intypes=="ee" )
                    for ( int i=0; i<n; ++i )
                        (*(func_e_ee)(f))( ret->v[i], ret->dv[i],
                                           pa[0]->to_r(i), pa[0]->to_dr(i),
                                           pa[1]->to_r(i), pa[1]->to_dr(i) );
                else if  ( tf->intypes=="eee" )
                    for ( int i=0; i<n; ++i )
                        (*(func_e_eee)(f))( ret->v[i], ret->dv[i],
                             pa[0]->to_r(i), pa[0]->to_dr(i),
                             pa[1]->to_r(i), pa[1]->to_dr(i),
                             pa[2]->to_r(i), pa[2]->to_dr(i) );
                else
                    throw S("BUG: unexpected intypes");
                return ret;
            } else
                throw S("BUG: unexpected outtype");
        }
        throw S("BUG: unexpected case");
    } catch( string& ex ) {
        throw ex+" in NodeFun "+tree_info()+" {intypes='"+basetypes+"', " +
            ( tf ? ( "cast to '"+tf->intypes+"', outtype='"+tf->outtype+"', " ) : "" ) +
            "context: "+ctx.context_info()+"}";
    }
}


void CNodeFun1::set_coord( CCoord& ret, int k ) const {
    CCoord r;
    arg[0]->set_coord( r, k );
    ret = fu->coord( &r );
}
void CNodeFun2::set_coord( CCoord& ret, int k ) const {
    CCoord r[2];
    arg[0]->set_coord( r[0], k );
    arg[1]->set_coord( r[1], k );
    ret = fu->coord( r+0, r+1 );
}
void CNodeFun3::set_coord( CCoord& ret, int k ) const {
    CCoord r[3];
    arg[0]->set_coord( r[0], k );
    arg[1]->set_coord( r[1], k );
    arg[2]->set_coord( r[2], k );
    ret = fu->coord( r+0, r+1, r+2 );
}

//! Returns string representation of this node.

string CNodeFun1::tree_info() const
{
    return fu->tag + "(" + arg[0]->tree_info() + ")";
}

//! Returns string representation of this node.

string CNodeFun2::tree_info() const
{
    return fu->tag + "(" + arg[0]->tree_info() + "," + arg[1]->tree_info() + ")";
}

//! Returns string representation of this node.

string CNodeFun3::tree_info() const
{
    return fu->tag + "(" +
        arg[0]->tree_info() + "," +
        arg[1]->tree_info() + "," +
        arg[2]->tree_info() + ")";
}


//**************************************************************************************************
//* CNodeRange: range node
//**************************************************************************************************


//! Evaluates the range in context, and returns a vector object pointer.

const RObj CNodeRange::tree_val( const CContext& ctx ) const
{
     // evaluate arguments, determine types, assert they are scalar.
    vector<RObj> pa(narg);
    bool all_int = true;
    for ( int iarg=0; iarg<narg; ++iarg ) {
        if ( arg[iarg] )
            pa[iarg] = arg[iarg]->tree_val( ctx );
        else if ( iarg==0 )
            pa[iarg] = RObj( new CObjInt( 0 ) ); // default begin
        else if ( iarg==1 )
            pa[iarg] = RObj( new CObjInt( ctx.nv-1 ) ); // default end
        else if ( iarg==2 )
            pa[iarg] = RObj( new CObjInt( 1 ) ); // default step
        if( pa[iarg]->base_type()!='i' )
            all_int = false;
        if ( pa[iarg]->is_vec() )
            throw "vectorial argument in range " + tree_info();
    }

    if ( all_int ) {
        PObjVecInt ret( new CObjVecInt );
        for ( int i = pa[0]->to_i();
              ( inclusive_end ? i<=pa[1]->to_i() : i<pa[1]->to_i() );
               i+=pa[2]->to_i() )
            ret->v.push_back( i );
        return ret;
    } else {
        PObjVecDbl ret( new CObjVecDbl );
        for ( double r = pa[0]->to_r();
              ( inclusive_end ? r<=pa[1]->to_r() : r<pa[1]->to_r() );
              r+=pa[2]->to_r() )
            ret->v.push_back( r );
        return ret;
    }
}

void CNodeRange::set_coord( CCoord& ret, int k ) const {
    CCoord r[3];
    arg[0]->set_coord( r[0], k );
    arg[1]->set_coord( r[1], k );
    arg[2]->set_coord( r[2], k );
    ret.name = "range";
}

//! Returns string representation of this node.

string CNodeRange::tree_info() const
{
    return "range(" + (arg[0] ? arg[0]->tree_info() : "0") + "," +
        (arg[1] ? arg[1]->tree_info() : "end") + "," +
        (arg[2] ? arg[2]->tree_info() : "1" ) + ")";
}


//**************************************************************************************************
//* CNodeList
//**************************************************************************************************


CNodeList::CNodeList( RNode _incremand, RNode _increment )
    : CNodeWithArgs(1)
{
    if ( RNodeList old = PCAST<const CNodeList>(_incremand) ) {
        arg = old->args();
    } else {
        arg[0] = _incremand;
    }
    if ( _increment )
        arg.push_back( _increment );
    narg = arg.size();
}


//! Evaluates the range in context, and returns a vector object pointer.

const RObj CNodeList::tree_val( const CContext& ctx ) const
{
     // evaluate arguments, determine types
    vector<RObj> pa(narg);
    bool all_int = true;
    bool all_dbl = true;
    bool has_enu = false;
    for ( int iarg=0; iarg<narg; ++iarg ) {
        if ( arg[iarg] )
            pa[iarg] = arg[iarg]->tree_val( ctx );
        char t = pa[iarg]->base_type();
        if( t!='i' )
            all_int = false;
        if( t!='i' && t!='d' && t!='e' )
            all_dbl = false;
        if( t=='e' )
            has_enu = true;
    }

    if ( all_int ) {
        PObjVecInt ret( new CObjVecInt );
        for ( int iarg=0; iarg<narg; ++iarg ) {
            if ( RObjVecInt rav = PCAST<const CObjVecInt>(pa[iarg]) )
                for ( int i=0; i<rav->size(); ++i )
                    ret->v.push_back( rav->to_i(i) );
            else
                ret->v.push_back( pa[iarg]->to_i() );
        }
        return ret;
    } else if ( all_dbl ) {
        if ( has_enu ) {
            PObjVecEnu ret( new CObjVecEnu );
            for ( int iarg=0; iarg<narg; ++iarg ) {
                if ( RObjVecNum rav = PCAST<const CObjVecNum>(pa[iarg]) ) {
                    for ( int i=0; i<rav->size(); ++i ) {
                        ret->v.push_back( rav->to_r(i) );
                        ret->dv.push_back( rav->to_dr(i) );
                    }
                } else {
                    ret->v.push_back( pa[iarg]->to_r() );
                    ret->dv.push_back( pa[iarg]->to_dr() );
                }
            }
            return ret;
        } else {
            PObjVecDbl ret( new CObjVecDbl );
            for ( int iarg=0; iarg<narg; ++iarg ) {
                if ( RObjVecNum rav = PCAST<const CObjVecNum>(pa[iarg]) )
                    for ( int i=0; i<rav->size(); ++i )
                        ret->v.push_back( rav->to_r(i) );
                else
                    ret->v.push_back( pa[iarg]->to_r() );
            }
            return ret;
        }
    } else {
        throw S("list-of-any not yet implemented");
    }
}

void CNodeList::set_coord( CCoord& ret, int k ) const {
    CCoord r[3];
    arg[0]->set_coord( r[0], k );
    arg[1]->set_coord( r[1], k );
    arg[2]->set_coord( r[2], k );
    ret.name = "list";
}

//! Returns string representation of this node.

string CNodeList::tree_info() const
{
    string ret = "list(";
    for ( int iarg=0; iarg<narg; ++iarg ) {
        ret += arg[iarg]->tree_info();
        if ( iarg<narg-1 )
            ret += ",";
    }
    return ret;           
}


//**************************************************************************************************
//* CNodeGeni: generalized integral node
//**************************************************************************************************


CNodeGeni::CNodeGeni( const class CGeni *_geni, RRef _ref )
    : CNodeWithArgs(_geni->narg), geni(_geni)
{
    arg.resize(narg);
    for ( int iarg=0; iarg<narg; ++iarg )
        arg[iarg] = geni->default_arg[iarg];
}

CNodeGeni::CNodeGeni( const class CGeni *_geni, RNode _a0 )
    : CNodeWithArgs(1), geni(_geni)
{
    if ( geni->narg!=1 )
        throw "syntax error: generalized integral '" + geni->name + "' needs " +
            S(geni->narg) + " arguments, not 1";
    arg[0] = _a0;
}

CNodeGeni::CNodeGeni( const class CGeni *_geni, RNode _a0, RNode _a1 )
    : CNodeWithArgs(2), geni(_geni)
{
    if ( geni->narg!=2 )
        throw "syntax error: generalized integral '" + geni->name + "' needs " +
            S(geni->narg) + " arguments, not 2";
    arg[0] = _a0;
    arg[1] = _a1;
}

//! A functional of 1,2,.. arguments (including operators like +,-,.. ).

const RObj CNodeGeni::tree_val( const CContext& ctx ) const
{
    try {
        if( ctx.dim!=CContext::_1 )
            throw "Unexpected is_vec() {context: " + ctx.context_info() +"}";
        CContext myctx(ctx);
        myctx.dim = CContext::_VI;
        myctx.nv = -1;
        // Evaluate arguments, and determine size of is_vec() arguments.
        vector<RObj> pa(narg);
        vector<RObjVecNum> pav(narg);
        int n=0;
        for ( int iarg=0; iarg<narg; ++iarg ) {
            pa[iarg] = arg[iarg]->tree_val( myctx );
            pav[iarg] = PCAST<const CObjVecNum>(pa[iarg]);
            if( !pav[iarg] )
                throw "argument " + S(iarg) + " of generalized integral " +
                    tree_info() + " is scalar: " + pa[iarg]->result_info();
            if ( n==0 )
                n = pav[iarg]->size();
            else if ( pav[iarg]->size() != n )
                throw S("vector arguments have different size");
        }
        // Now evaluate the function.
        double val;
        geni->eval( &val, NULL, n, pav ); // TODO: compute errors
        return RObjDbl( new CObjDbl( val ) );

    } catch( string& ex ) {
        throw ex+" in NodeGeni "+tree_info()+" {context: "+ctx.context_info()+"}";
    }
}

void CNodeGeni::set_coord( CCoord& ret, int k ) const {
    string rnam = geni->name+"(";
    vector<string> argunit;
    for ( int iarg=0; iarg<narg; ++iarg ){
        CCoord carg;
        arg[iarg]->set_coord( carg, k );
        argunit.push_back( carg.unit );
        rnam += carg.name;
        rnam += iarg==narg-1 ? ")" : ",";
    }
    string runit;
    if        ( geni->unit_hint == "?" ) {
        runit = "";
        cout << "TODO: hard-code unit for this generalized integral\n";
    } else if ( geni->unit_hint == "-" ) {
        runit = "";
    } else if ( geni->unit_hint == "a" ) {
        if ( narg<1 )
            throw "BUG: unit hint '" + geni->unit_hint + "'  inconsistent with narg=" + S(narg);
        runit = argunit[0];
    } else if ( geni->unit_hint == "ab" ) {
        if ( narg<2 )
            throw "BUG: unit hint '" + geni->unit_hint + "'  inconsistent with narg=" + S(narg);
        runit = argunit[0] + "*" + argunit[1];
    } else {
        throw  "BUG: unforeseen unit hint '" + geni->unit_hint + "'";
    }
    
    ret = CCoord( rnam, runit );
}


//! Returns string representation of this node.

string CNodeGeni::tree_info() const
{
    string ret = geni->name+"(";
    for ( int iarg=0; iarg<narg; ++iarg )
        ret += arg[iarg]->tree_info() + (iarg<narg-1 ? "," : ")" );
    return ret;
}


//**************************************************************************************************
//* CNodeCvin: curve functional node
//**************************************************************************************************


CNodeCvin::CNodeCvin( const class CCvin *_cvin, RRef _ref, RNode _a0, RNode _a1 )
    : CNodeWithArgs(2), cvin(_cvin), ref(_ref)
{
    if ( cvin->narg!=2 )
        throw "syntax error: curve functional '" + cvin->txt + "' needs " +
            S(cvin->narg) + " arguments, not 2";
    arg[0] = _a0;
    arg[1] = _a1;
}

const RObj CNodeCvin::tree_val( const CContext& ctx ) const
{
    try {
        if( ctx.dim!=CContext::_1 )
            throw "Unexpected is_vec() {context: " + ctx.context_info() +"}";
        CContext myctx(ctx);
        myctx.dim = CContext::_1;
        // Evaluate arguments.
        vector<double> a(narg);
        for ( int iarg=0; iarg<narg; ++iarg ) {
            RObj pa = arg[iarg]->tree_val( myctx );
            try {
                a[iarg] = pa->to_r();
            } catch( string& ex ) {
                throw "argument " + S(iarg) + " of curve functional " + tree_info() +
                    " is not a scalar number: " + pa->result_info() + " => " + ex;
            }
        }
        int k = ref->get_k( ctx );
        ROlo f = NOlm::mem_get(k);
        ROlc fc = R2C( f );
        if ( !fc )
            throw S("This is not a curve file => cannot integrate");
        int j = ref->get_j( ctx, f->nJ() );
        // now evaluate the function
        double val;
        cvin->eval( &val, NULL, fc, k, j, a ); // TODO: compute errors?
        return PObjDbl( new CObjDbl( val ) );
    } catch( string& ex ) {
        throw ex+" in NodeCvin "+tree_info()+" {context: "+ctx.context_info()+"}";
    }
}

void CNodeCvin::set_coord( CCoord& ret, int k ) const {
//*
//*     CCoord r[maxarg];
//*     for ( int iarg=0; iarg<narg; ++iarg )
//*         arg[iarg]->set_coord( r[iarg], k );
//*     if      ( const CFunc1* fu = dynamic_cast<const CFunc1*>( fun ) )
//*         ret = fu->coord( r+0);
//*     else if ( const CFunc2* fu = dynamic_cast<const CFunc2*>( fun ) )
//*         ret = fu->coord( r+0, r+1 );
//*     else if ( const CFunc3* fu = dynamic_cast<const CFunc3*>( fun ) )
//*         ret = fu->coord( r+0, r+1, r+2 );
}

//! Returns string representation of this node.

string CNodeCvin::tree_info() const
{
    string ret = cvin->txt+"(..)";
    for ( int iarg=0; iarg<narg; ++iarg )
        ret += arg[iarg]->tree_info() + (iarg<narg-1 ? "," : ")" );
    return ret;
}


//**************************************************************************************************
//* CNodeVal: end node containing a numeric value
//**************************************************************************************************

CNodeVal::CNodeVal( double _val ) : val( PObjDbl( new CObjDbl( _val ) ) ) {}

CNodeVal::CNodeVal( int _val ) : val( PObjInt( new CObjInt( _val ) ) ) {}

bool CNodeVal::looks_like_indices() const
{
    return PCAST<const CObjInt>(val)!=NULL;
}

const RObj CNodeVal::tree_val( const CContext& ctx ) const
{
    return val;
}

void CNodeVal::set_coord( CCoord& ret, int k ) const
{
    ret = CCoord(val->to_s(), "");
}

string CNodeVal::tree_info() const
{
    return "val[" + val->to_s() + "]";
}


//**************************************************************************************************
//* CNodeIva: indexed variable node
//**************************************************************************************************

void CNodeIva::npar_exec( int *np ) const
{
    if ( var->typ == CVar::_CP )
        *np = max( *np, var->num+1 );
}

//! Evaluate reference in given context. Result returned in ret->

const RObj CNodeIva::tree_val( const CContext& ctx ) const
{
    try {
        if ( (var->typ == CVar::_I || var->typ == CVar::_J ||
              var->typ == CVar::_K) && ( ref->ti || ref->tj || ref->tk ) )
            throw S("invalid cross reference: index cannot be indexed");

        // Get k from context:
        int k = ref->get_k( ctx );

        // Return k-dependent reference (unless already done above):
        if ( var->typ == CVar::_K ) {
            return RObjInt( new CObjInt( k ) );
        }

        // Proceed with file reference f[k]:
        POlo f = NOlm::mem_get(k);

        if        ( var->typ == CVar::_NK ) {
            return RObjInt( new CObjInt( NOlm::mem_size() ) );
        } else if ( var->typ == CVar::_NJ ) {
            return RObjInt( new CObjInt( f->nJ() ) );
        } else if ( var->typ == CVar::_R ) {
            if ( var->num>= f->RPar.size() )
                throw "invalid reference " + var->var_info();
            return RObjDbl( new CObjDbl( f->RPar[var->num].val ) ); // TODO: RPar -> Enu? Obj?
        }

        // Return k-j-dependent reference (unless done above):
        if        ( var->typ == CVar::_J ) {
            if ( ref->ti )
                throw S("J does not require 3 indices");
            return RObjInt( new CObjInt( ctx.j ) );
        } else if ( var->typ == CVar::_Z ) {
            if ( ref->ti )
                throw S("Z does not require 3 indices");
            if ( var->num>= f->nZ() )
                throw "invalid reference " + var->var_info();
            int j = ref->get_j( ctx, f->nJ() );
            return f->V[j]->z[var->num];
        }

        // References for data file
        POld fd = P2D( f );
        if ( fd ) {
            if        ( var->pointwise() ) {
                if        ( ctx.dim==CContext::_1 ) {
                    int j = ref->get_j( ctx, f->nJ() );
                    PSpec sj = fd->VS(j);
                    int i;
                    if ( ref->ti ) {
                        i = ref->ti->tree_val_idx( ctx, "I" );
                    } else {
                        i = ctx.i;
                        if ( i==(int)-1 )
                            throw S("missing index I");
                    }
                    if( i>= sj->size() )
                        throw "i="+S(i)+" while there are only ni="+S(sj->size())+" points";
                    if       ( var->typ == CVar::_I ) {
                        return RObjInt( new CObjInt( i ) );
                    } else if ( var->typ == CVar::_X ) {
                        return RObjDbl( new CObjDbl( sj->x[i] ) );
                    } else if ( var->typ == CVar::_Y ) {
                        if ( ctx.want_error && sj->has_dy() )
                            return RObjEnu( new CObjEnu( sj->y[i], sj->dy[i] ) );
                        else
                            return RObjDbl( new CObjDbl( sj->y[i] ) );
                    } else if ( var->typ == CVar::_DY ) {
                        return RObjDbl( new CObjDbl( sj->dy[i] ) );
                    } else
                        throw S("BUG: unexpected case");

                } else if ( ctx.dim==CContext::_VI ) {
                    if ( ref->ti ) {
                        CContext myctx = ctx;
                        myctx.dim = CContext::_1;
                        if        ( var->typ == CVar::_I ) {
                            PObjVecInt pret( new CObjVecInt( ctx.nv ) );
                            for ( int ii=0; ii<ctx.nv; ++ii ) {
                                myctx.i = ii;
                                int i = ref->ti->tree_val_idx( myctx, "I" );
                                pret->v[ii] = i;
                            }
                            return pret;
                        } else if ( var->typ == CVar::_Y ) {
                            if ( ctx.want_error ) {
                                PObjVecEnu pret( new CObjVecEnu( ctx.nv ) );
                                for ( int ii=0; ii<ctx.nv; ++ii ) {
                                    myctx.i = ii;
                                    int j = ref->get_j( myctx, f->nJ() );
                                    RSpec sj = fd->VS(j);
                                    if ( !sj->has_dy() )
                                        goto vi_y_dbl;
                                    int i = ref->ti->tree_val_idx( myctx, "I" );
                                    if( i >= sj->size() )
                                        throw "i="+S(i)+" while there are only"
                                            " ni="+S(sj->size())+" points";
                                    pret->v[ii] = sj->y[i];
                                    pret->dv[ii] = sj->dy[i];
                                }
                                return pret;
                            }
                        vi_y_dbl:
                            PObjVecDbl pret2( new CObjVecDbl( ctx.nv ) );
                            for ( int ii=0; ii<ctx.nv; ++ii ) {
                                myctx.i = ii;
                                int j = ref->get_j( myctx, f->nJ() );
                                RSpec sj = fd->VS(j);
                                int i = ref->ti->tree_val_idx( myctx, "I" );
                                if( i >= sj->size() )
                                    throw "i="+S(i)+" while there are only"
                                        " ni="+S(sj->size())+" points";
                                pret2->v[ii] = sj->y[i];
                            }
                            return pret2;
                        } else if ( var->typ == CVar::_X ) {
                            PObjVecDbl pret( new CObjVecDbl( ctx.nv ) );
                            for ( int ii=0; ii<ctx.nv; ++ii ) {
                                myctx.i = ii;
                                int j = ref->get_j( myctx, f->nJ() );
                                RSpec sj = fd->VS(j);
                                int i = ref->ti->tree_val_idx( myctx, "I" );
                                if( i >= sj->size() )
                                    throw "i="+S(i)+" while there are only"
                                        " ni="+S(sj->size())+" points";
                                pret->v[ii] = sj->x[i];
                            }
                            return pret;
                        } else if ( var->typ == CVar::_DY ) {
                            PObjVecDbl pret( new CObjVecDbl( ctx.nv ) );
                            for ( int ii=0; ii<ctx.nv; ++ii ) {
                                myctx.i = ii;
                                int j = ref->get_j( myctx, f->nJ() );
                                RSpec sj = fd->VS(j);
                                int i = ref->ti->tree_val_idx( myctx, "I" );
                                if( i >= sj->size() )
                                    throw "i="+S(i)+" while there are only"
                                        " ni="+S(sj->size())+" points";
                                pret->v[ii] = sj->dy[i];
                            }
                            return pret;
                        }
                        throw S("BUG: unexpected case");

                    } else {
                        int j = ref->get_j( ctx, f->nJ() );
                        PSpec sj = fd->VS(j);
                        int nv = ctx.nv;
                        if ( nv==-1 ) {
                            nv = sj->size();
                        } else if ( nv != sj->size() )
                            throw "ref_val "+var->var_info()+": requested vec("+S(ctx.nv)+
                                "), found vec("+S( sj->size() );
                        if        ( var->typ == CVar::_I ) {
                            PObjVecInt pret( new CObjVecInt( nv ) );
                            for( int i=0; i<nv; ++i )
                                pret->v[i] = i;
                            return pret;
                        } else if ( var->typ == CVar::_X ) {
                            return RObjVecDbl( new CObjVecDbl( sj->x ) );
                        } else if ( var->typ == CVar::_Y ) {
                            if ( ctx.want_error && sj->has_dy() )
                                return RObjVecEnu( new CObjVecEnu( sj->y, sj->dy ) );
                            else
                                return RObjVecDbl( new CObjVecDbl( sj->y ) );
                        } else if ( var->typ == CVar::_DY ) {
                            return RObjVecDbl( new CObjVecDbl( sj->dy ) );
                        }
                        throw S("BUG: unexpected case");
                    }
                }
                throw S("BUG: invalid context::dim");

            } else if ( var->typ == CVar::_NI ) {
                int j = ref->get_j( ctx, f->nJ() );
                PSpec sj = fd->VS(j);
                return PObjInt( new CObjInt( sj->size() ) );

            } else
                throw "invalid ref("+var->var_info()+") in data file";

        } else { // References for curve file.
            POlc fc = P2C( f );
            if ( fc ) {
                if ( ref->ti )
                    throw S("unexpected i dependence of curve property");
                int j = ref->get_j( ctx, f->nJ() );
                PCurve cj = fc->VC(j);
                if       ( var->typ == CVar::_CP ) {
                    if ( var->num>= fc->nP )
                        throw "invalid p ref("+var->var_info()+") in curve file";
                    return RObjDbl( new CObjDbl( cj->P[var->num] ) );

                } else if ( var->typ == CVar::_COUTCOME ) {
                    return RObjInt( new CObjInt( cj->fitOutcome ) );

                } else if ( var->typ == CVar::_CCHI2 ) {
                    return RObjDbl( new CObjDbl( cj->fitChi2 ) );

                } else if ( var->typ == CVar::_CR2 ) {
                    return RObjDbl( new CObjDbl( cj->fitR2 ) );

                }
                throw "invalid ref("+var->var_info()+") in curve file";
            }
        }
    } catch( string& ex ) {
        throw ex+" in NodeIva "+tree_info()+" {context: "+ctx.context_info()+"}";
    }
    return NULL;
}

//! Description of reference (for use in ??)

void CNodeIva::set_coord( CCoord& ret, int k_in ) const
{
    if        ( var->typ == CVar::_K ) {
        ret = CCoord("k","");
        return;
    } else if ( var->typ == CVar::_J ) {
        ret = CCoord("j","");
        return;
    } else if ( var->typ == CVar::_I ) {
        ret = CCoord("i","");
        return;
    }

    int k;
    if (ref->tk) {
        // get k from reference :
        CContext ctx( k_in );
        k = ref->tk->tree_val_idx( ctx, "K" );
    } else {
        k = k_in;
    }
    POlo f = NOlm::mem_get(k);

    if  ( var->typ == CVar::_Z ) {
        if (var->num>=f->ZCo.size())
            throw "set_coord: z" + S(var->num) + " undefined";
        ret = f->ZCo[var->num];
    } else if  ( var->typ == CVar::_R ) {
        if (var->num>=f->RPar.size())
            throw "set_coord: r" + S(var->num) + " undefined";
        ret = f->RPar[var->num].Co;
    } else if ( var->typ == CVar::_NJ ) {
        ret = CCoord("#spectra", "");
    } else {
        POld fd = P2D( f );
        POlc fc = P2C( f );
        if         ( fd ) {
            if       ( var->typ == CVar::_X )
                ret = fd->xco;
            else if  ( var->typ == CVar::_Y )
                ret = fd->yco;
            else if  ( var->typ == CVar::_DY )
                ret = CCoord( "d"+fd->yco.name, fd->yco.unit );
            else if  ( var->typ == CVar::_NI )
                ret = CCoord("#points", "");
            else
                throw "data file has no " + var->var_info();
        } else if ( fc ) {
            if     ( var->typ == CVar::_CP ) {
                if( var->num>=fc->nP )
                    throw "invalid reference p" + S(var->num);
                ret = fc->PCo[var->num];
            }
            else if ( var->typ == CVar::_COUTCOME )
                ret = CCoord("fit_outcome", "");
            else if ( var->typ == CVar::_CCHI2 )
                ret = CCoord("chi2", "");
            else if ( var->typ == CVar::_CR2 )
                ret = CCoord("R2", "");
            else if ( var->typ == CVar::_C )
                ret = CCoord(fc->expr,"");
            else
                throw "curve file has no " + var->var_info();
        } else
            throw S("BUG: ref->set_coord unexpected else");
    }
}

//! Returns string representation of this node.

string CNodeIva::tree_info() const
{
    return "ref[" + var->var_info() + "]";
}


//**************************************************************************************************
//* CNodeIdf: identifier node
//**************************************************************************************************

CNodeIdf::CNodeIdf( string _idf )
    : idf( _idf ) {}

const RObj CNodeIdf::tree_val( const CContext& ctx ) const
{
    PObj ret = GReg->find( idf );
    if ( !ret )
        throw "uninitialized variable "+idf;
    return ret;
}

void CNodeIdf::set_coord( CCoord& ret, int k ) const
{
    ret = CCoord( idf, "" );  // primitiver gehts nicht
}


//**************************************************************************************************
//* CNodeDummy: function argument (t) node
//**************************************************************************************************

const RObj CNodeDummy::tree_val( const CContext& ctx ) const
{
    if ( !(ctx.vt) )
        throw "BUG: tree_val with *vt=0";
    RObjVecDbl pret( new CObjVecDbl( *(ctx.vt) ) );
    return pret;
}

//**************************************************************************************************
//* CNodeCev: curve evaluation node
//**************************************************************************************************

CNodeCev::CNodeCev( RRef& _ref, RNode _arg )
    : ref( _ref ), arg( _arg )
{
    if( ref->ti )
        throw S("curve reference must not depend on index i");
}

//! A curve evaluation: fc[k,j](arg).

const RObj CNodeCev::tree_val( const CContext& ctx ) const
{
    try {
        // get k, curve, j from context:
        int k = ref->get_k( ctx );
        ROlc fc = NOlm::mem_get_C( k );
        int j = ref->get_j( ctx, fc->nJ() );

        // evaluate function argument:
        RObj farg = arg->tree_val( ctx );

        // evaluate the curve:
        if ( RObjVecDbl pv_farg = PCAST<const CObjVecDbl>(farg) ) {
            return fc->eval_curve( pv_farg->v, k, j, ctx.want_error );
        } else if ( RObjNum pn_farg = PCAST<const CObjNum>(farg) ) {
            return fc->eval_curve_scalar( pn_farg->to_r(), k, j, ctx.want_error );
        } else
            throw S("unexpected argument type");
    } catch( string& ex ) {
        throw ex + " in NodeCev " + tree_info() + " {context: " + ctx.context_info() + "}";
    }
}

void CNodeCev::set_coord( CCoord& ret, int k ) const {
    ret = CCoord( "c", "" ); } // primitiver gehts nicht


//**************************************************************************************************
//* CNodeMixin: Base class for mixing in data from a "convolution" file
//**************************************************************************************************

const RObj CNodeMixin::tree_val( const CContext& ctx ) const
{
    // Shift:
    RObj res_shift = shift->tree_val( ctx );
    RObjNum par = PCAST<const CObjNum>(res_shift);
    if( !par )
        throw S("shift is not a scalar");
    double theshift = par->to_r();

    // Resolution is given by context 'ctx':
    int kconv, jconv;
    PSpec sv;
    NCurveFile::get_conv( &sv, &kconv, &jconv, ctx.k, ctx.j );

    PObjVecDbl pret( ctx.want_error ? (new CObjVecDbl) : (new CObjVecEnu) );

    if ( kconv==-1 ) {
        return copy_theory( ctx, theshift );
    } else {
        double conv_norm; // normalization constant (integral of resolution)
        double conv_step; // step in x/t, if equidistant
        string info = "resolution " + S(kconv) + ", spectrum " + S(jconv);
        conv_norm = sv->norm( info );
        conv_step = sv->has_step( ctx.vt_step ) ? ctx.vt_step : 0;
        if( !conv_step && !allow_slow_conv )
            throw S("grid is not equidistant; slow convolution is not allowed");
        return convolve( ctx, theshift, sv, conv_norm, conv_step );
    }
}


//**************************************************************************************************
//* CNodeConvBase: Base class for convoluting a function with a spectrum.
//**************************************************************************************************

void CNodeConvBase::kdep_exec( bool *kd ) const
{
    theory->kdep_exec( kd );
    shift->kdep_exec( kd );
}

void CNodeConvBase::npar_exec( int *np ) const
{
    theory->npar_exec( np );
    shift->npar_exec( np );
}


//**************************************************************************************************
//* CNodeConv: convoluted function node
//**************************************************************************************************

RObj CNodeConv::copy_theory( const CContext& ctx, double theshift ) const
{
    return theory->tree_val( ctx );
}

RObjVecDbl CNodeConv::convolve( const CContext& ctx, double theshift, const PSpec& sv,
                                double conv_norm, double conv_step ) const
{
    int n = ctx.vt->size();
    int nv = sv->size();
    CContext cv_ctx(ctx);
    vector<double> tt; // theory grid
    cv_ctx.vt = &tt;
    PObjVecDbl ret;
    PObjVecEnu dret;
    if ( ctx.want_error ) {
       dret = PObjVecEnu( new CObjVecEnu(n) );
       ret = dret;
    } else {
        ret = PObjVecDbl( new CObjVecDbl(n) );
        dret = NULL;
    }
    // Convolute theory with resolution.
    if( conv_step ){ // accelerated version with equidistant grids
        // precompute theory on unified grid
        int ntt = n + nv - 1;
        tt.resize(ntt);
        double tt0 = (*(ctx.vt))[0] - sv->x[0] - theshift - (nv-1)*conv_step;
        for ( int ii=0; ii<ntt; ++ii ) // ii=(nv-1)+i-iv
            tt[ii] = tt0 + ii*conv_step;
        RObj res_theory = theory->tree_val( cv_ctx );
        // now convolve
        for ( int i=0; i<n; ++i ){
            ret->v[i] = 0;
            if( dret )
                dret->dv[i] = 0;
            for ( int iv=0; iv<nv; ++iv ) {
                ret->v[i] += res_theory->to_r(nv-1+i-iv) * sv->y[iv];
                if( dret )
                    dret->dv[i] += fabs(res_theory->to_r(nv-1+i-iv)) * SQR(sv->dy[iv]);
            }
            ret->v[i] *= conv_step / conv_norm;
            if( dret )
                dret->dv[i] = sqrt(dret->dv[i]) * conv_step / conv_norm;
        }
    } else { // simplest version: non-equidistant x_out and x_res.
        throw S("BROKEN CODE (tt never used)");
        ret->v.assign( n, 0. );
        if( dret )
            dret->dv.assign( n, 0. );
        tt.resize(n);
        for ( int iv=0; iv<nv; ++iv ) {
            double dx; // variable step
            if      ( iv==0 )
                dx = sv->x[1] - sv->x[0];
            else if ( iv==nv-1 )
                dx = sv->x[nv-1] - sv->x[nv-2];
            else
                dx = ( sv->x[iv+1] - sv->x[iv-1] ) / 2;
            for ( int i=0; i<n; ++i )
                tt[i] = (*(ctx.vt))[i] - sv->x[iv] - theshift;
            RObj res_theory = theory->tree_val( cv_ctx );
            for ( int i=0; i<n; ++i ) {
                ret->v[i] += res_theory->to_r(i) * sv->y[iv] * dx;
                if( dret )
                    dret->dv[i] += res_theory->to_r(i) * SQR(sv->dy[iv]);
            }
        }
        for ( int i=0; i<n; ++i ) {
            ret->v[i] /= conv_norm;
            if( dret )
                dret->dv[i] = sqrt(dret->dv[i])/conv_norm;
        }
    }
    return ret;
}


//**************************************************************************************************
//* CNodePConv: convolution with primitive
//**************************************************************************************************

RObj CNodePConv::copy_theory( const CContext& ctx, double theshift ) const
{
    throw "non-convolution not implemented for pconv";
    // to implement this efficiently, I would like to have CObj::operator+
    // for which we should go to shared pointers ...
}

RObjVecDbl CNodePConv::convolve( const CContext& ctx, double theshift, const PSpec& sv,
                                 double conv_norm, double conv_step ) const
{
    if( conv_norm <= 0 )
        throw "BUG: pconv called with conv_norm="+S(conv_norm);
    int n = ctx.vt->size();
    int nv = sv->size();
    CContext cv_ctx(ctx);
    vector<double> tt;
    cv_ctx.vt = &tt;
    PObjVecDbl ret;
    PObjVecEnu dret;
    if ( ctx.want_error ) {
       dret = PObjVecEnu( new CObjVecEnu(n) );
       ret = dret;
       // A check. We don't care about the cost O(nv), since this node is O(n*nv).
       for ( int iv=0; iv<nv; ++iv ) {
           if ( !isfinite(sv->dy[iv]) )
               throw S("resolution function contains non-finite error");
           if ( sv->dy[iv]<0 )
               throw S("resolution function contains negative error");
       }
    } else {
        ret = PObjVecDbl( new CObjVecDbl(n) );
        dret = NULL;
    }
    // Convolute theory with resolution.
    if( conv_step ){ // accelerated version with equidistant grids
        // precompute primitive on unified grid
        int ntt = n + nv + 1;
        tt.resize(ntt);
        double tt0 = (*(ctx.vt))[0] - sv->x[0] - theshift - (nv-0.5)*conv_step;
        for ( int ii=0; ii<ntt; ++ii ) // ii=(nv-1)+i-iv
            tt[ii] = tt0 + ii*conv_step;
        RObj res_theory = theory->tree_val( cv_ctx );
        if ( res_theory->size() != ntt )
            throw S("BUG: unexpected size conflict");
        for ( int it=0; it<ntt; ++it )
            if ( !isfinite(res_theory->to_r(it)) )
                throw S("theory returns infinite value");
        for ( int it=0; it<ntt-1; ++it )
            if( res_theory->to_r(it) > res_theory->to_r(it+1) )
                throw "theory is decreasing from S(" + S(tt[it]) + "=" + S(res_theory->to_r(it)) +
                    " to S(" + S(tt[it+1]) + "=" + S(res_theory->to_r(it+1));
        // now convolve resolution with differential of primitive
        for ( int i=0; i<n; ++i ){
            ret->v[i] = 0;
            if( dret )
                dret->dv[i] = 0;
            for ( int iv=0; iv<nv; ++iv ) {
                double igral = res_theory->to_r(nv+i-iv) - res_theory->to_r(nv-1+i-iv);
                ret->v[i] += igral * sv->y[iv];
                if( dret )
                    dret->dv[i] += igral * SQR(sv->dy[iv]);
            }
            ret->v[i] /= conv_norm;
            if ( !isfinite(ret->v[i]) )
                throw S("BUG: pconv result has nonfinite result");
            if( dret ) {
                if ( !isfinite(dret->dv[i]) )
                    throw S("BUG: pconv result has nonfinite variance");
                if ( dret->dv[i]<0 )
                    throw S("BUG: pconv result has negative variance");
                dret->dv[i] = sqrt(dret->dv[i]) / conv_norm;
                if ( !isfinite(dret->dv[i]) )
                    throw S("BUG: pconv result has nonfinite error");
            }
        }
    } else { // simplest version: non-equidistant x_out and x_res.
        throw S("PCONV BROKEN");
/*
        ret->v.assign( n, 0. );
        if( with_error )
            ret->dv.assign( n, 0. );
        tt.resize(n);
        RObj res_theory_low;
        RObj res_theory_hig;
        double dx;
        for ( int iv=0; iv<nv; ++iv ) {
            if      ( iv==0 )
                dx = sv->x[1] - sv->x[0];
            else if ( iv==nv-1 )
                dx = sv->x[nv-1] - sv->x[nv-2];
            else
                dx = ( sv->x[iv+1] - sv->x[iv-1] ) / 2;
            for ( int i=0; i<n; ++i ) {
                tt[i] = (*(ctx.vt))[i] - sv->x[iv] - theshift - dx/2;
                theory->tree_val( res_theory_low, cv_ctx );
            }
            for ( int i=0; i<n; ++i ) {
                tt[i] = (*(ctx.vt))[i] - sv->x[iv] - theshift + dx/2;
                theory->tree_val( res_theory_hig, cv_ctx );
            }
            for ( int i=0; i<n; ++i ) {
                double igral = res_theory_hig->to_r(i)-res_theory_low->to_r(i);
                ret->v[i] += igral * sv->y[iv] * dx;
                if ( with_error )
                    ret->dv[i] += igral * SQR(sv->dy[iv]) * dx;
            }
        }
        for ( int i=0; i<n; ++i ) {
            ret->v[i] /= conv_norm;
            if ( with_error )
                ret->dv[i] = sqrt(ret->dv[i]) / conv_norm;
        }
*/
    }
    return ret;
}


//**************************************************************************************************
//* CNodeDirac: node that copies the resolution function
//**************************************************************************************************

void CNodeDirac::kdep_exec( bool *kd ) const
{
    shift->kdep_exec( kd );
}

void CNodeDirac::npar_exec( int *np ) const
{
    shift->npar_exec( np );
}

RObj CNodeDirac::copy_theory( const CContext& ctx, double theshift ) const
{
    // Naive representation of delta function,
    // for use when convolution is switched off.
    PObjVecDbl ret( new CObjVecDbl( ctx.vt->size()));
    for ( int i=0; i<ctx.vt->size(); ++i )
        ret->v[i] = fabs((*(ctx.vt))[i]-theshift)<1e-2 ? 1e2 : 0;
    return ret;
}

//! Convolution or Dirac function: conv(f, shift), dirac(shift).

RObjVecDbl CNodeDirac::convolve( const CContext& ctx, double theshift, const PSpec& sv,
                                 double conv_norm, double conv_step ) const
{
    int n = ctx.vt->size();
    PObjVecDbl ret;
    PObjVecEnu dret;
    if ( ctx.want_error ) {
        dret = PObjVecEnu( new CObjVecEnu(n) );
        ret = dret;
    } else {
        ret = PObjVecDbl( new CObjVecDbl(n) );
        dret = NULL;
    }
    // Delta function copies resolution.
    vector<double> tt(n), yy(n), dd(n);
    // shift in t
    for ( int i=0; i<n; ++i )
        tt[i] = (*(ctx.vt))[i]-theshift;
    // interpolation
    try {
        sv->intpol( tt, &yy, dret ? &dd : NULL );
    } catch( string& s ) {
        throw "error while evaluating delta function with shift "+S(theshift)+": "+s;
    }
    // normalization
    for ( int i=0; i<n; ++i ) {
        ret->v[i] = yy[i] / conv_norm;
        if( dret )
            dret->dv[i] = dd[i] / conv_norm;
    }
    return ret;
}

void CNodeDirac::set_coord( CCoord& ret, int k ) const
{
    ret = CCoord("resol", "");
}
