//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  obj.cpp
//! \brief Data containers: CObj and its children.

#include <string>
#include <vector>
#include <cmath>
#include <boost/format.hpp>

using namespace std;
using boost::format;

#include "defs.hpp"
#include "ptr.hpp"
#include "obj.hpp"


//**************************************************************************************************
//  Static methods
//**************************************************************************************************

PObj CObj::objectify( int val ) { return PObjInt( new CObjInt(val) ); }
PObj CObj::objectify( double val ) { return PObjDbl( new CObjDbl(val) ); }
PObj CObj::objectify( double val, double err ) { return PObjEnu( new CObjEnu(val, err) ); }
PObj CObj::objectify( string val ) { return PObjStr( new CObjStr(val) ); }


//**************************************************************************************************
//  CObjInt
//**************************************************************************************************

//! Returns string representation.

string CObjInt::to_s( int maxlen, int minlen, int prec ) const
{
    return S(val);
}


//**************************************************************************************************
//  CObjDbl
//**************************************************************************************************

//! Returns string representation.

string CObjDbl::to_s( int maxlen, int minlen, int prec ) const
{
    string form = str( format( "%%#%i.%ig" ) % maxlen % prec );
    string ret = str( format( form ) % val );
    return ret;
}

//! Returns textual representation of the class instance, for debugging.

string CObjDbl::result_info() const
{
    if( ::isnan(val) )
        return "CObjDbl(no value assigned)";
    else
        return "CObjDbl("+to_s()+")";
}


//**************************************************************************************************
//  CObjEnu
//**************************************************************************************************

//! Returns string representation.

string CObjEnu::to_s( int maxlen, int minlen, int prec ) const
{
    return S(val) + "+-" + S(err);
}

//! Returns textual representation of the class instance, for debugging.

string CObjEnu::result_info() const
{
    if( ::isnan(val) )
        return "CObjEnu(no value assigned)";
    else
        return "CObjDbl("+to_s()+")";
}


//**************************************************************************************************
//  CObjVec
//**************************************************************************************************

//! Asserts that vector has exactly one element.

void CObjVec::assert_is_lonely() const
{
    if ( size()!=1 )
        throw "BUG: vector has "+S(size())+" elements while exactly one was expected";
}


//**************************************************************************************************
//  CObjVecInt
//**************************************************************************************************

//! Returns string representation.

string CObjVecInt::to_s( int maxlen, int minlen, int prec ) const
{
    string ret;
    for ( int i=0; i<v.size(); ++i ) {
        ret += S(v[i]);
        if ( i<v.size()-1 )
            ret += ",";
    }
    return ret;
}


//! Returns textual representation of the class instance, for debugging.

string CObjVecInt::result_info() const
{
    return to_s();
}


//! Returns pointer to new scalar object initialized with the only element of this vector.

PObj CObjVecInt::unpack_lonely_element() const
{
    assert_is_lonely();
    return PObjInt( new CObjInt( v[0] ) );
}


//**************************************************************************************************
//  CObjVecDbl
//**************************************************************************************************

//! Returns string representation.

string CObjVecDbl::to_s( int maxlen, int minlen, int prec ) const
{
    string ret;
    for ( int i=0; i<v.size(); ++i ) {
        ret += S(v[i]);
        if ( i<v.size()-1 )
            ret += ",";
    }
    return ret;
}


//! Returns textual representation of the class instance, for debugging.

string CObjVecDbl::result_info() const
{
    return to_s();
}


//! Returns pointer to new scalar object initialized with the only element of this vector.

PObj CObjVecDbl::unpack_lonely_element() const
{
    assert_is_lonely();
    return PObjDbl( new CObjDbl( v[0] ) );
}


//**************************************************************************************************
//  CObjVecEnu
//**************************************************************************************************


//! Returns string representation.

string CObjVecEnu::to_s( int maxlen, int minlen, int prec ) const
{
    return "["+S(size())+" floating-point entries with errors]";
}

//! Returns textual representation of the class instance, for debugging.

string CObjVecEnu::result_info() const
{
    return to_s();
}


//! Returns pointer to new scalar object initialized with the only element of this vector.

PObj CObjVecEnu::unpack_lonely_element() const
{
    assert_is_lonely();
    return PObjEnu( new CObjEnu( v[0], dv[0] ) );
}
